from m.core import Good, Issue, OneOf, issue, one_of, subprocess


def get_branch() -> OneOf[Issue, str]:
    """Get the current git branch name.

    Returns:
        A `OneOf` containing an `Issue` or a string specifying the branch.
    """
    return subprocess.eval_cmd('git rev-parse --abbrev-ref HEAD')


def stage_all() -> OneOf[Issue, str]:
    """Stage the current changes in the branch.

    Returns:
        A `OneOf` containing an `Issue` or the response from `git add .`.
    """
    return subprocess.eval_cmd('git add .').flat_map_bad(
        lambda err: issue('git add failure', cause=err),
    )


def commit(msg: str) -> OneOf[Issue, str]:
    """Create a commit.

    Args:
        msg: commit description.

    Returns:
        A `OneOf` containing an `Issue` or the response from the git command.
    """
    return subprocess.eval_cmd(f'git commit -m "{msg}"').flat_map_bad(
        lambda err: issue('git commit failure', cause=err),
    )


def push_branch(branch: str) -> OneOf[Issue, str]:
    """Push branch.

    Args:
        branch: name of branch to push.

    Returns:
        A `OneOf` containing an `Issue` or the response from the command.
    """
    return subprocess.eval_cmd(f'git push -u origin "{branch}"').flat_map_bad(
        lambda err: issue('git push failure', cause=err),
    )


def stash() -> OneOf[Issue, str]:
    """Stash the current changes in the branch.

    Returns:
        A `OneOf` containing an `Issue` or the response from `git stash`.
    """
    return subprocess.eval_cmd('git stash')


def stash_pop() -> OneOf[Issue, str]:
    """Pop the changes stored in the git stash.

    Returns:
        A `OneOf` containing an `Issue` or the response from `git stash pop`.
    """
    return subprocess.eval_cmd('git stash pop')


def checkout_branch(branch: str, create: bool = True) -> OneOf[Issue, str]:
    """Checkout a branch.

    Args:
        branch: name of branch to checkout
        create: create new branch

    Returns:
        A `OneOf` containing an `Issue` of the git response.
    """
    opt = '-b' if create else ''
    return subprocess.eval_cmd(f'git checkout {opt} {branch}').flat_map_bad(
        lambda err: issue('git checkout failure', cause=err),
    )


def get_first_commit_sha() -> OneOf[Issue, str]:
    """Find the first commit sha in the repository.

    Returns:
        A `OneOf` containing an `Issue` or a string of the first commit sha.
    """
    return subprocess.eval_cmd('git rev-list --max-parents=0 HEAD')


def get_current_commit_sha() -> OneOf[Issue, str]:
    """Find the sha of the current commit.

    Returns:
        A `OneOf` containing an `Issue` or a string of the current commit sha.
    """
    return subprocess.eval_cmd('git rev-parse HEAD')


def get_remote_url() -> OneOf[Issue, str]:
    """Find the remote url of the repo.

    Returns:
        A `OneOf` containing an `Issue` or a string with the url.
    """
    return subprocess.eval_cmd('git config --get remote.origin.url')


def get_commits(
    first: str,
    latest: str = 'HEAD',
) -> OneOf[Issue, list[str] | None]:
    """Get a list of all the commits between two tags.

    May return `None` in the special case when `first` is `0.0.0`.

    Args:
        first: The first tag.
        latest: The second tag, defaults to HEAD.

    Returns:
        A `OneOf` containing an `Issue` or a list of all the commits.
    """
    if first == '0.0.0':
        return Good(None)
    cmd = f'git log {latest}...{first} --oneline --no-color'
    return subprocess.eval_cmd(cmd).map(lambda out: out.splitlines())


def _extract_status(msg: str) -> tuple[str, str]:
    matches = [
        ('Untracked files', 'untracked'),
        ('Your branch is ahead', 'ahead'),
        ('Your branch is behind', 'behind'),
        ('working tree clean', 'clean'),
        ('Changes to be committed', 'staged'),
        ('Changed but not updated', 'dirty'),
        ('Changes not staged', 'dirty'),
        ('Unmerged paths', 'dirty'),
        ('diverged', 'diverged'),
    ]
    for substr, key in matches:
        if substr in msg:
            return key, substr
    return '?', 'unknown'


def get_status() -> OneOf[Issue, tuple[str, str]]:
    """Find the current git status.

    Returns:
        A `OneOf` containing an `Issue` or a word denoting the git status.
    """
    res = subprocess.eval_cmd('git status')
    return one_of(
        lambda: [
            _extract_status(msg)
            for msg in res
        ],
    )


def raw_status() -> OneOf[Issue, str]:
    """Obtain the output of "git status".

    Returns:
        A `OneOf` containing an `Issue` or the output of "git status".
    """
    return subprocess.eval_cmd('git status')



from m.cli import add_arg, command, run_main
from m.core.io import env
from pydantic import BaseModel, Field


class Arguments(BaseModel):
    """End the release process.

    Merges the release/hotfix branch into the master branch.
    """

    github_token: str = Field(
        proxy=add_arg(
            '--github-token',
            default=env('GITHUB_TOKEN'),
            help='Github PAT (default: env.GITHUB_TOKEN)',
        ),
    )


@command(
    name='end_release',
    help='end the release process',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.end_release import end_release

    return run_main(lambda: end_release(arg.github_token))



from m.cli import command
from pydantic import BaseModel, Field


class Arguments(BaseModel):
    """Open a block to group several messages in the build log."""

    name: str = Field(
        description='block name to open',
        positional=True,
        required=True,
    )
    description: str = Field(
        description='block description',
        positional=True,
        required=True,
    )


@command(
    name='open',
    help='open a block',
    model=Arguments,
)
def run(arg: Arguments):
    from m.log import Logger

    logger = Logger('m.cli')
    logger.open_block(arg.name, arg.description)
    return 0



from m.cli import add_arg, command, run_main
from m.core.io import env
from pydantic import BaseModel, Field


class Arguments(BaseModel):
    """Create Github pull request(s) to review the release."""

    github_token: str = Field(
        proxy=add_arg(
            '--github-token',
            default=env('GITHUB_TOKEN'),
            help='Github PAT (default: env.GITHUB_TOKEN)',
        ),
    )


@command(
    name='review_release',
    help='create release pull requests',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.review_release import review_release

    return run_main(lambda: review_release(arg.github_token))



from m.cli import command, run_main
from pydantic import BaseModel


class Arguments(BaseModel):
    """Display a single word representing the current git status.

    example::

        $ m git status
        clean

    statuses::

        unknown
        untracked
        stash
        clean
        ahead
        behind
        staged
        dirty
        diverged
        ?
    """


def _get_status(status_desc: tuple[str, str]) -> str:
    return status_desc[0]


@command(
    name='status',
    help='display the current git status',
    model=Arguments,
)
def run():
    from m import git
    return run_main(
        lambda: git.get_status().map(_get_status),
        print,
    )



import argparse
import logging

from m.cli import run_cli
from m.log import logging_config
from m.version import VERSION


def main_args(parser: argparse.ArgumentParser) -> None:
    """Handle an ArgumentParser instance to add global cli arguments.

    Args:
        parser: The Argument parser instance.
    """
    parser.add_argument('--version', action='version', version=VERSION)


def main():
    """Execute the cli."""
    logging_config(logging.NOTSET)
    run_cli(__file__, main_args)


if __name__ == '__main__':  # pragma: no cover
    main()



from re import sub

from pydantic import BaseModel


def to_camel(snake_case: str) -> str:
    """Transform a string in snake_case to camel case.

    Args:
        snake_case: string in snake case.

    Returns:
        string in camel case.
    """
    s = sub('(_|-)+', ' ', snake_case).title().replace(' ', '')
    return ''.join([s[0].lower(), s[1:]])


class CamelModel(BaseModel):
    """Allows models to be defined with camel case properties.

    See:
        https://medium.com/analytics-vidhya/camel-case-models-with-fast-api-and-pydantic-5a8acb6c0eee
    """

    class Config:
        """Config to allow camel case properties."""

        alias_generator = to_camel
        allow_population_by_field_name = True
